/* imposter (OO.org Impress viewer)
** Copyright (C) 2003 Gurer Ozen <madcat@e-kolay.net>
** This code is free software; you can redistribute it and/or
** modify it under the terms of GNU General Public License.
*/

#include "common.h"

struct fs_struct {
	GtkWidget *win;
	void (*callback)(const char *filename);
};

static void
fs_ok (GtkWidget *w, struct fs_struct *fs)
{
	char *file;

	file = (char *)gtk_file_selection_get_filename (GTK_FILE_SELECTION (fs->win));
	fs->callback (file);
	gtk_widget_destroy (fs->win);
	free (fs);
}

static void
fs_cancel (GtkWidget *w, struct fs_struct *fs)
{
	gtk_widget_destroy (fs->win);
	free (fs);
}

static gint
fs_delete (GtkWidget *w, GdkEvent *ev, struct fs_struct *fs)
{
	free (fs);
	return FALSE;
}

void
select_file (void (*func)(const char *filename), char *title)
{
	struct fs_struct *fs;
	GtkWidget *win, *b;

	fs = malloc (sizeof (struct fs_struct));
	fs->callback = func;
	win = gtk_file_selection_new (title);
	fs->win = win;
	gtk_container_set_border_width (GTK_CONTAINER (win), 10);
	b = GTK_FILE_SELECTION (win)->ok_button;
	g_signal_connect (G_OBJECT (b), "clicked", G_CALLBACK (fs_ok), fs);
	b = GTK_FILE_SELECTION (win)->cancel_button;
	g_signal_connect (G_OBJECT (b), "clicked", G_CALLBACK (fs_cancel), fs);
	g_signal_connect (G_OBJECT (win), "delete_event", G_CALLBACK (fs_delete), fs);
	gtk_widget_show (win);
}

void
draw_bezier (GdkDrawable *d, GdkGC *gc, int x0, int y0, int x1, int y1, int x2, int y2, int x3, int y3)
{
	int x, y, nx, ny;
	int ax, bx, cx, ay, by, cy;
	double t, t2, t3;

	x = x0;
	y = y0;

	cx = 3 * (x1 - x0);
	bx = 3 * (x2 - x1) - cx;
	ax = x3 - x0 - cx - bx;
	cy = 3 * (y1 - y0);
	by = 3 * (y2 - y1) - cy;
	ay = y3 - y0 - cy - by;

	for (t = 0; t < 1; t += 0.01) {
		t2 = t * t;
		t3 = t2 * t;
		nx = ax * t3 + bx * t2 + cx * t + x0;
		ny = ay * t3 + by * t2 + cy * t + y0;
		gdk_draw_line (d, gc, x, y, nx, ny);
		x = nx;
		y = ny;
	}
}
