/* imposter (OO.org Impress viewer)
** Copyright (C) 2003-2004 Gurer Ozen <madcat@e-kolay.net>
** This code is free software; you can redistribute it and/or
** modify it under the terms of GNU General Public License.
*/

#include "common.h"
#include "render_ctx.h"

#include <math.h>

void
r_draw_pixbuf (render_ctx *ctx, char *name, int x, int y, int w, int h)
{
	GdkPixbuf *orig_pb, *pb;

	if (name[0] == '#') name++;

	orig_pb = oo_doc_get_gfx (ctx->doc, name);
	pb = gdk_pixbuf_scale_simple (orig_pb, w, h, GDK_INTERP_BILINEAR);
	gdk_draw_pixbuf (ctx->d, ctx->gc, pb, 0, 0, ctx->start_x + x, ctx->start_y + y, w, h, GDK_RGB_DITHER_NONE, 0, 0);
	g_object_unref (pb);
}

void
r_tile_pixbuf (render_ctx *ctx, char *name, int x, int y, int w, int h)
{
	GdkPixbuf *orig_pb, *pb;
	int gw, gh;
	int gx, gy;

	if (name[0] == '#') name++;

	orig_pb = oo_doc_get_gfx (ctx->doc, name);
	gw = gdk_pixbuf_get_width (orig_pb);
	gh = gdk_pixbuf_get_height (orig_pb);
	pb = orig_pb;
//	pb = gdk_pixbuf_scale_simple (orig_pb, gw, gh, GDK_INTERP_BILINEAR);

	for (gx = x; gx < w; gx += gw) {
		for (gy = y; gy < h; gy += gh) {
			gdk_draw_pixbuf (ctx->d, ctx->gc, pb, 0, 0,
				ctx->start_x + gx, ctx->start_y + gy, gw, gh, GDK_RGB_DITHER_NONE, 0, 0);
		}
	}

//	g_object_unref (pb);
}

void
r_draw_move (render_ctx *ctx, int x, int y)
{
	ctx->pen_x = x;
	ctx->pen_y = y;
}

void
r_draw_lineto (render_ctx *ctx, int x, int y)
{
	gdk_draw_line (ctx->d, ctx->gc,
		ctx->start_x + ctx->pen_x, ctx->start_y + ctx->pen_y,
		ctx->start_x + x, ctx->start_y + y);
	ctx->pen_x = x;
	ctx->pen_y = y;
}

void
r_draw_curveto (render_ctx *ctx, int x, int y, int c1x, int c1y, int c2x, int c2y)
{
	draw_bezier (ctx->d, ctx->gc,
		ctx->start_x + ctx->pen_x, ctx->start_y + ctx->pen_y,
		ctx->start_x + c1x, ctx->start_y + c1y,
		ctx->start_x + c2x, ctx->start_y + c2y,
		ctx->start_x + x, ctx->start_y + y);
	ctx->pen_x = x;
	ctx->pen_y = y;
}

void
r_draw_rect (render_ctx *ctx, gboolean fill, int x, int y, int w, int h, int roundness)
{
	int a;

	if (0 == roundness) {
		gdk_draw_rectangle (ctx->d, ctx->gc, fill,
			ctx->start_x + x, ctx->start_y + y,
			w, h);
		return;
	}
	gdk_draw_arc (ctx->d, ctx->gc, fill,
		ctx->start_x + x, ctx->start_y + y,
		roundness, roundness,
		90 * 64, 90 * 64);
	gdk_draw_arc (ctx->d, ctx->gc, fill,
		ctx->start_x + x + w - roundness, ctx->start_y + y,
		roundness, roundness,
		0, 90 * 64);
	gdk_draw_arc (ctx->d, ctx->gc, fill,
		ctx->start_x + x + w - roundness, ctx->start_y + y + h - roundness,
		roundness, roundness,
		270 * 64, 90 * 64);
	gdk_draw_arc (ctx->d, ctx->gc, fill,
		ctx->start_x + x, ctx->start_y + y + h - roundness,
		roundness, roundness,
		180 * 64, 90 * 64);

	a = roundness / 2;
	if (fill) {
		gdk_draw_rectangle (ctx->d, ctx->gc, TRUE,
			ctx->start_x + x + a, ctx->start_y + y,
			w - a -a, h);
		gdk_draw_rectangle (ctx->d, ctx->gc, TRUE,
			ctx->start_x + x, ctx->start_y + y + a,
			w, h - a - a);
		return;
	}
	gdk_draw_line (ctx->d, ctx->gc,
		ctx->start_x + x + a, ctx->start_y + y,
		ctx->start_x + x + w - a, ctx->start_y + y);
	gdk_draw_line (ctx->d, ctx->gc,
		ctx->start_x + x + a, ctx->start_y + y + h,
		ctx->start_x + x + w - a, ctx->start_y + y + h);
	gdk_draw_line (ctx->d, ctx->gc,
		ctx->start_x + x, ctx->start_y + y + a,
		ctx->start_x + x, ctx->start_y + y + h - a);
	gdk_draw_line (ctx->d, ctx->gc,
		ctx->start_x + x + w, ctx->start_y + y + a,
		ctx->start_x + x + w, ctx->start_y + y + h - a);
}

 static int tantable[3600] = { 
      0, 1, 3 , 5, 6, 8, 10, 12, 13, 15, 
      17, 19, 20 , 22, 24, 26, 27, 29, 31, 33, 
      34, 36, 38 , 40, 41, 43, 45, 47, 48, 50, 
      52, 54, 55 , 57, 59, 61, 62, 64, 66, 68, 
      69, 71, 73 , 75, 76, 78, 80, 82, 83, 85, 
      87, 89, 91 , 92, 94, 96, 98, 99, 101, 103, 
      105, 106, 108 , 110, 112, 113, 115, 117, 119, 121, 
      122, 124, 126 , 128, 129, 131, 133, 135, 136, 138, 
      140, 142, 144 , 145, 147, 149, 151, 153, 154, 156, 
      158, 160, 161 , 163, 165, 167, 169, 170, 172, 174, 
      176, 178, 179 , 181, 183, 185, 187, 188, 190, 192, 
      194, 196, 198 , 199, 201, 203, 205, 207, 208, 210, 
      212, 214, 216 , 218, 219, 221, 223, 225, 227, 229, 
      230, 232, 234 , 236, 238, 240, 241, 243, 245, 247, 
      249, 251, 253 , 254, 256, 258, 260, 262, 264, 266, 
      267, 269, 271 , 273, 275, 277, 279, 281, 282, 284, 
      286, 288, 290 , 292, 294, 296, 298, 300, 301, 303, 
      305, 307, 309 , 311, 313, 315, 317, 319, 321, 322, 
      324, 326, 328 , 330, 332, 334, 336, 338, 340, 342, 
      344, 346, 348 , 350, 352, 354, 356, 358, 360, 361, 
      363, 365, 367 , 369, 371, 373, 375, 377, 379, 381, 
      383, 385, 387 , 389, 391, 393, 395, 397, 399, 401, 
      404, 406, 408 , 410, 412, 414, 416, 418, 420, 422, 
      424, 426, 428 , 430, 432, 434, 436, 438, 441, 443, 
      445, 447, 449 , 451, 453, 455, 457, 459, 462, 464, 
      466, 468, 470 , 472, 474, 476, 479, 481, 483, 485, 
      487, 489, 492 , 494, 496, 498, 500, 502, 505, 507, 
      509, 511, 513 , 516, 518, 520, 522, 525, 527, 529, 
      531, 533, 536 , 538, 540, 542, 545, 547, 549, 552, 
      554, 556, 558 , 561, 563, 565, 568, 570, 572, 575, 
      577, 579, 582 , 584, 586, 589, 591, 593, 596, 598, 
      600, 603, 605 , 608, 610, 612, 615, 617, 620, 622, 
      624, 627, 629 , 632, 634, 637, 639, 641, 644, 646, 
      649, 651, 654 , 656, 659, 661, 664, 666, 669, 671, 
      674, 677, 679 , 682, 684, 687, 689, 692, 695, 697, 
      700, 702, 705 , 708, 710, 713, 715, 718, 721, 723, 
      726, 729, 731 , 734, 737, 739, 742, 745, 748, 750, 
      753, 756, 759 , 761, 764, 767, 770, 772, 775, 778, 
      781, 784, 786 , 789, 792, 795, 798, 801, 804, 806, 
      809, 812, 815 , 818, 821, 824, 827, 830, 833, 836, 
      839, 842, 845 , 848, 851, 854, 857, 860, 863, 866, 
      869, 872, 875 , 878, 881, 884, 887, 890, 894, 897, 
      900, 903, 906 , 909, 913, 916, 919, 922, 926, 929, 
      932, 935, 939 , 942, 945, 948, 952, 955, 958, 962, 
      965, 969, 972 , 975, 979, 982, 986, 989, 993, 996, 
      999, 1003, 1007 , 1010, 1014, 1017, 1021, 1024, 1028, 1031, 
      1035, 1039, 1042 , 1046, 1050, 1053, 1057, 1061, 1064, 1068, 
      1072, 1076, 1079 , 1083, 1087, 1091, 1095, 1098, 1102, 1106, 
      1110, 1114, 1118 , 1122, 1126, 1130, 1134, 1138, 1142, 1146, 
      1150, 1154, 1158 , 1162, 1166, 1170, 1174, 1179, 1183, 1187, 
      1191, 1195, 1200 , 1204, 1208, 1213, 1217, 1221, 1226, 1230, 
      1234, 1239, 1243 , 1248, 1252, 1257, 1261, 1266, 1270, 1275, 
      1279, 1284, 1289 , 1293, 1298, 1303, 1307, 1312, 1317, 1322, 
      1327, 1331, 1336 , 1341, 1346, 1351, 1356, 1361, 1366, 1371, 
      1376, 1381, 1386 , 1391, 1396, 1401, 1407, 1412, 1417, 1422, 
      1428, 1433, 1438 , 1444, 1449, 1455, 1460, 1465, 1471, 1476, 
      1482, 1488, 1493 , 1499, 1505, 1510, 1516, 1522, 1528, 1533, 
      1539, 1545, 1551 , 1557, 1563, 1569, 1575, 1581, 1587, 1594, 
      1600, 1606, 1612 , 1619, 1625, 1631, 1638, 1644, 1651, 1657, 
      1664, 1670, 1677 , 1684, 1690, 1697, 1704, 1711, 1718, 1725, 
      1732, 1739, 1746 , 1753, 1760, 1767, 1774, 1781, 1789, 1796, 
      1804, 1811, 1818 , 1826, 1834, 1841, 1849, 1857, 1864, 1872, 
      1880, 1888, 1896 , 1904, 1912, 1920, 1929, 1937, 1945, 1954, 
      1962, 1971, 1979 , 1988, 1996, 2005, 2014, 2023, 2032, 2041, 
      2050, 2059, 2068 , 2077, 2087, 2096, 2105, 2115, 2125, 2134, 
      2144, 2154, 2164 , 2174, 2184, 2194, 2204, 2214, 2225, 2235, 
      2246, 2256, 2267 , 2278, 2288, 2299, 2310, 2321, 2333, 2344, 
      2355, 2367, 2378 , 2390, 2402, 2414, 2426, 2438, 2450, 2462, 
      2475, 2487, 2500 , 2512, 2525, 2538, 2551, 2564, 2578, 2591, 
      2605, 2618, 2632 , 2646, 2660, 2674, 2688, 2703, 2717, 2732, 
      2747, 2762, 2777 , 2792, 2808, 2823, 2839, 2855, 2871, 2887, 
      2904, 2920, 2937 , 2954, 2971, 2988, 3006, 3023, 3041, 3059, 
      3077, 3096, 3114 , 3133, 3152, 3171, 3190, 3210, 3230, 3250, 
      3270, 3291, 3312 , 3333, 3354, 3375, 3397, 3419, 3442, 3464, 
      3487, 3510, 3533 , 3557, 3581, 3605, 3630, 3655, 3680, 3706, 
      3732, 3758, 3784 , 3811, 3839, 3866, 3894, 3923, 3951, 3981, 
      4010, 4040, 4071 , 4102, 4133, 4165, 4197, 4230, 4263, 4297, 
      4331, 4366, 4401 , 4437, 4473, 4510, 4548, 4586, 4625, 4664, 
      4704, 4745, 4786 , 4828, 4871, 4915, 4959, 5004, 5050, 5097, 
      5144, 5192, 5242 , 5292, 5343, 5395, 5448, 5502, 5557, 5613, 
      5671, 5729, 5789 , 5850, 5912, 5975, 6040, 6106, 6174, 6243, 
      6313, 6385, 6459 , 6534, 6612, 6691, 6771, 6854, 6939, 7026, 
      7115, 7206, 7300 , 7396, 7494, 7595, 7699, 7806, 7915, 8028, 
      8144, 8263, 8386 , 8512, 8642, 8776, 8915, 9057, 9205, 9357, 
      9514, 9676, 9844 , 10018, 10198, 10385, 10578, 10779, 10987, 11204, 
      11429, 11664, 11908 , 12163, 12428, 12705, 12995, 13299, 13617, 13950, 
      14300, 14668, 15055 , 15463, 15894, 16349, 16831, 17342, 17885, 18464, 
      19080, 19739, 20445 , 21204, 22021, 22903, 23858, 24897, 26029, 27270, 
      28635, 30143, 31819 , 33692, 35798, 38186, 40915, 44063, 47736, 52077, 
      57285, 63651, 71608 , 81838, 95477, 114571, 143210, 190935, 286369, 572522, 
      753695995, -573393, -286586 , -191032, -143264, -114606, -95501, -81856, -71621, -63662, 
      -57294, -52084, -47742 , -44068, -40919, -38190, -35802, -33695, -31821, -30145, 
      -28637, -27272, -26031 , -24898, -23860, -22904, -22022, -21205, -20447, -19740, 
      -19081, -18464, -17886 , -17343, -16832, -16350, -15894, -15464, -15056, -14668, 
      -14300, -13950, -13617 , -13299, -12996, -12706, -12429, -12163, -11908, -11664, 
      -11430, -11204, -10988 , -10779, -10579, -10385, -10198, -10018, -9844, -9676, 
      -9514, -9357, -9205 , -9058, -8915, -8776, -8642, -8512, -8386, -8263, 
      -8144, -8028, -7915 , -7806, -7699, -7595, -7494, -7396, -7300, -7206, 
      -7115, -7026, -6939 , -6854, -6772, -6691, -6612, -6535, -6459, -6385, 
      -6313, -6243, -6174 , -6106, -6040, -5975, -5912, -5850, -5789, -5729, 
      -5671, -5614, -5557 , -5502, -5448, -5395, -5343, -5292, -5242, -5192, 
      -5144, -5097, -5050 , -5004, -4959, -4915, -4871, -4828, -4786, -4745, 
      -4704, -4664, -4625 , -4586, -4548, -4510, -4473, -4437, -4401, -4366, 
      -4331, -4297, -4263 , -4230, -4197, -4165, -4133, -4102, -4071, -4040, 
      -4010, -3981, -3951 , -3923, -3894, -3866, -3839, -3811, -3784, -3758, 
      -3732, -3706, -3680 , -3655, -3630, -3605, -3581, -3557, -3533, -3510, 
      -3487, -3464, -3442 , -3419, -3397, -3375, -3354, -3333, -3312, -3291, 
      -3270, -3250, -3230 , -3210, -3191, -3171, -3152, -3133, -3114, -3096, 
      -3077, -3059, -3041 , -3023, -3006, -2988, -2971, -2954, -2937, -2920, 
      -2904, -2887, -2871 , -2855, -2839, -2823, -2808, -2792, -2777, -2762, 
      -2747, -2732, -2717 , -2703, -2688, -2674, -2660, -2646, -2632, -2618, 
      -2605, -2591, -2578 , -2564, -2551, -2538, -2525, -2512, -2500, -2487, 
      -2475, -2462, -2450 , -2438, -2426, -2414, -2402, -2390, -2378, -2367, 
      -2355, -2344, -2333 , -2321, -2310, -2299, -2288, -2278, -2267, -2256, 
      -2246, -2235, -2225 , -2214, -2204, -2194, -2184, -2174, -2164, -2154, 
      -2144, -2134, -2125 , -2115, -2106, -2096, -2087, -2077, -2068, -2059, 
      -2050, -2041, -2032 , -2023, -2014, -2005, -1996, -1988, -1979, -1971, 
      -1962, -1954, -1945 , -1937, -1929, -1920, -1912, -1904, -1896, -1888, 
      -1880, -1872, -1864 , -1857, -1849, -1841, -1834, -1826, -1819, -1811, 
      -1804, -1796, -1789 , -1781, -1774, -1767, -1760, -1753, -1746, -1739, 
      -1732, -1725, -1718 , -1711, -1704, -1697, -1690, -1684, -1677, -1670, 
      -1664, -1657, -1651 , -1644, -1638, -1631, -1625, -1619, -1612, -1606, 
      -1600, -1594, -1587 , -1581, -1575, -1569, -1563, -1557, -1551, -1545, 
      -1539, -1534, -1528 , -1522, -1516, -1510, -1505, -1499, -1493, -1488, 
      -1482, -1476, -1471 , -1465, -1460, -1455, -1449, -1444, -1438, -1433, 
      -1428, -1422, -1417 , -1412, -1407, -1401, -1396, -1391, -1386, -1381, 
      -1376, -1371, -1366 , -1361, -1356, -1351, -1346, -1341, -1336, -1331, 
      -1327, -1322, -1317 , -1312, -1307, -1303, -1298, -1293, -1289, -1284, 
      -1279, -1275, -1270 , -1266, -1261, -1257, -1252, -1248, -1243, -1239, 
      -1234, -1230, -1226 , -1221, -1217, -1213, -1208, -1204, -1200, -1195, 
      -1191, -1187, -1183 , -1179, -1175, -1170, -1166, -1162, -1158, -1154, 
      -1150, -1146, -1142 , -1138, -1134, -1130, -1126, -1122, -1118, -1114, 
      -1110, -1106, -1102 , -1098, -1095, -1091, -1087, -1083, -1079, -1076, 
      -1072, -1068, -1064 , -1061, -1057, -1053, -1050, -1046, -1042, -1039, 
      -1035, -1031, -1028 , -1024, -1021, -1017, -1014, -1010, -1007, -1003, 
      -1000, -996, -993 , -989, -986, -982, -979, -975, -972, -969, 
      -965, -962, -958 , -955, -952, -948, -945, -942, -939, -935, 
      -932, -929, -926 , -922, -919, -916, -913, -909, -906, -903, 
      -900, -897, -894 , -890, -887, -884, -881, -878, -875, -872, 
      -869, -866, -863 , -860, -857, -854, -851, -848, -845, -842, 
      -839, -836, -833 , -830, -827, -824, -821, -818, -815, -812, 
      -809, -806, -804 , -801, -798, -795, -792, -789, -786, -784, 
      -781, -778, -775 , -772, -770, -767, -764, -761, -759, -756, 
      -753, -750, -748 , -745, -742, -739, -737, -734, -731, -729, 
      -726, -723, -721 , -718, -715, -713, -710, -708, -705, -702, 
      -700, -697, -695 , -692, -689, -687, -684, -682, -679, -677, 
      -674, -671, -669 , -666, -664, -661, -659, -656, -654, -651, 
      -649, -646, -644 , -641, -639, -637, -634, -632, -629, -627, 
      -624, -622, -620 , -617, -615, -612, -610, -608, -605, -603, 
      -600, -598, -596 , -593, -591, -589, -586, -584, -582, -579, 
      -577, -575, -572 , -570, -568, -565, -563, -561, -558, -556, 
      -554, -552, -549 , -547, -545, -542, -540, -538, -536, -533, 
      -531, -529, -527 , -525, -522, -520, -518, -516, -513, -511, 
      -509, -507, -505 , -502, -500, -498, -496, -494, -492, -489, 
      -487, -485, -483 , -481, -479, -476, -474, -472, -470, -468, 
      -466, -464, -462 , -459, -457, -455, -453, -451, -449, -447, 
      -445, -443, -441 , -438, -436, -434, -432, -430, -428, -426, 
      -424, -422, -420 , -418, -416, -414, -412, -410, -408, -406, 
      -404, -402, -399 , -397, -395, -393, -391, -389, -387, -385, 
      -383, -381, -379 , -377, -375, -373, -371, -369, -367, -365, 
      -363, -361, -360 , -358, -356, -354, -352, -350, -348, -346, 
      -344, -342, -340 , -338, -336, -334, -332, -330, -328, -326, 
      -324, -322, -321 , -319, -317, -315, -313, -311, -309, -307, 
      -305, -303, -301 , -300, -298, -296, -294, -292, -290, -288, 
      -286, -284, -282 , -281, -279, -277, -275, -273, -271, -269, 
      -267, -266, -264 , -262, -260, -258, -256, -254, -253, -251, 
      -249, -247, -245 , -243, -241, -240, -238, -236, -234, -232, 
      -230, -229, -227 , -225, -223, -221, -219, -218, -216, -214, 
      -212, -210, -208 , -207, -205, -203, -201, -199, -198, -196, 
      -194, -192, -190 , -188, -187, -185, -183, -181, -179, -178, 
      -176, -174, -172 , -170, -169, -167, -165, -163, -161, -160, 
      -158, -156, -154 , -153, -151, -149, -147, -145, -144, -142, 
      -140, -138, -136 , -135, -133, -131, -129, -128, -126, -124, 
      -122, -121, -119 , -117, -115, -113, -112, -110, -108, -106, 
      -105, -103, -101 , -99, -98, -96, -94, -92, -91, -89, 
      -87, -85, -83 , -82, -80, -78, -76, -75, -73, -71, 
      -69, -68, -66 , -64, -62, -61, -59, -57, -55, -54, 
      -52, -50, -48 , -47, -45, -43, -41, -40, -38, -36, 
      -34, -33, -31 , -29, -27, -26, -24, -22, -20, -19, 
      -17, -15, -13 , -12, -10, -8, -6, -5, -3, -1, 
      0, 1, 3 , 5, 6, 8, 10, 12, 13, 15, 
      17, 19, 20 , 22, 24, 26, 27, 29, 31, 33, 
      34, 36, 38 , 40, 41, 43, 45, 47, 48, 50, 
      52, 54, 55 , 57, 59, 61, 62, 64, 66, 68, 
      69, 71, 73 , 75, 76, 78, 80, 82, 83, 85, 
      87, 89, 91 , 92, 94, 96, 98, 99, 101, 103, 
      105, 106, 108 , 110, 112, 113, 115, 117, 119, 121, 
      122, 124, 126 , 128, 129, 131, 133, 135, 136, 138, 
      140, 142, 144 , 145, 147, 149, 151, 153, 154, 156, 
      158, 160, 161 , 163, 165, 167, 169, 170, 172, 174, 
      176, 178, 179 , 181, 183, 185, 187, 188, 190, 192, 
      194, 196, 198 , 199, 201, 203, 205, 207, 208, 210, 
      212, 214, 216 , 218, 219, 221, 223, 225, 227, 229, 
      230, 232, 234 , 236, 238, 240, 241, 243, 245, 247, 
      249, 251, 253 , 254, 256, 258, 260, 262, 264, 266, 
      267, 269, 271 , 273, 275, 277, 279, 281, 282, 284, 
      286, 288, 290 , 292, 294, 296, 298, 300, 301, 303, 
      305, 307, 309 , 311, 313, 315, 317, 319, 321, 322, 
      324, 326, 328 , 330, 332, 334, 336, 338, 340, 342, 
      344, 346, 348 , 350, 352, 354, 356, 358, 360, 361, 
      363, 365, 367 , 369, 371, 373, 375, 377, 379, 381, 
      383, 385, 387 , 389, 391, 393, 395, 397, 399, 401, 
      404, 406, 408 , 410, 412, 414, 416, 418, 420, 422, 
      424, 426, 428 , 430, 432, 434, 436, 438, 441, 443, 
      445, 447, 449 , 451, 453, 455, 457, 459, 462, 464, 
      466, 468, 470 , 472, 474, 476, 479, 481, 483, 485, 
      487, 489, 492 , 494, 496, 498, 500, 502, 505, 507, 
      509, 511, 513 , 516, 518, 520, 522, 525, 527, 529, 
      531, 533, 536 , 538, 540, 542, 545, 547, 549, 552, 
      554, 556, 558 , 561, 563, 565, 568, 570, 572, 575, 
      577, 579, 582 , 584, 586, 589, 591, 593, 596, 598, 
      600, 603, 605 , 608, 610, 612, 615, 617, 620, 622, 
      624, 627, 629 , 632, 634, 637, 639, 641, 644, 646, 
      649, 651, 654 , 656, 659, 661, 664, 666, 669, 671, 
      674, 677, 679 , 682, 684, 687, 689, 692, 695, 697, 
      700, 702, 705 , 708, 710, 713, 715, 718, 721, 723, 
      726, 729, 731 , 734, 737, 739, 742, 745, 748, 750, 
      753, 756, 759 , 761, 764, 767, 770, 772, 775, 778, 
      781, 784, 786 , 789, 792, 795, 798, 801, 804, 806, 
      809, 812, 815 , 818, 821, 824, 827, 830, 833, 836, 
      839, 842, 845 , 848, 851, 854, 857, 860, 863, 866, 
      869, 872, 875 , 878, 881, 884, 887, 890, 894, 897, 
      900, 903, 906 , 909, 913, 916, 919, 922, 926, 929, 
      932, 935, 939 , 942, 945, 948, 952, 955, 958, 962, 
      965, 969, 972 , 975, 979, 982, 986, 989, 993, 996, 
      999, 1003, 1006 , 1010, 1014, 1017, 1021, 1024, 1028, 1031, 
      1035, 1039, 1042 , 1046, 1050, 1053, 1057, 1061, 1064, 1068, 
      1072, 1076, 1079 , 1083, 1087, 1091, 1095, 1098, 1102, 1106, 
      1110, 1114, 1118 , 1122, 1126, 1130, 1134, 1138, 1142, 1146, 
      1150, 1154, 1158 , 1162, 1166, 1170, 1174, 1179, 1183, 1187, 
      1191, 1195, 1200 , 1204, 1208, 1213, 1217, 1221, 1226, 1230, 
      1234, 1239, 1243 , 1248, 1252, 1257, 1261, 1266, 1270, 1275, 
      1279, 1284, 1289 , 1293, 1298, 1303, 1307, 1312, 1317, 1322, 
      1327, 1331, 1336 , 1341, 1346, 1351, 1356, 1361, 1366, 1371, 
      1376, 1381, 1386 , 1391, 1396, 1401, 1407, 1412, 1417, 1422, 
      1428, 1433, 1438 , 1444, 1449, 1454, 1460, 1465, 1471, 1476, 
      1482, 1488, 1493 , 1499, 1505, 1510, 1516, 1522, 1528, 1533, 
      1539, 1545, 1551 , 1557, 1563, 1569, 1575, 1581, 1587, 1594, 
      1600, 1606, 1612 , 1619, 1625, 1631, 1638, 1644, 1651, 1657, 
      1664, 1670, 1677 , 1684, 1690, 1697, 1704, 1711, 1718, 1725, 
      1732, 1739, 1746 , 1753, 1760, 1767, 1774, 1781, 1789, 1796, 
      1804, 1811, 1818 , 1826, 1834, 1841, 1849, 1857, 1864, 1872, 
      1880, 1888, 1896 , 1904, 1912, 1920, 1929, 1937, 1945, 1954, 
      1962, 1971, 1979 , 1988, 1996, 2005, 2014, 2023, 2032, 2041, 
      2050, 2059, 2068 , 2077, 2087, 2096, 2105, 2115, 2125, 2134, 
      2144, 2154, 2164 , 2174, 2184, 2194, 2204, 2214, 2225, 2235, 
      2246, 2256, 2267 , 2278, 2288, 2299, 2310, 2321, 2333, 2344, 
      2355, 2367, 2378 , 2390, 2402, 2414, 2426, 2438, 2450, 2462, 
      2475, 2487, 2500 , 2512, 2525, 2538, 2551, 2564, 2578, 2591, 
      2605, 2618, 2632 , 2646, 2660, 2674, 2688, 2703, 2717, 2732, 
      2747, 2762, 2777 , 2792, 2808, 2823, 2839, 2855, 2871, 2887, 
      2904, 2920, 2937 , 2954, 2971, 2988, 3006, 3023, 3041, 3059, 
      3077, 3096, 3114 , 3133, 3152, 3171, 3190, 3210, 3230, 3250, 
      3270, 3291, 3312 , 3333, 3354, 3375, 3397, 3419, 3441, 3464, 
      3487, 3510, 3533 , 3557, 3581, 3605, 3630, 3655, 3680, 3706, 
      3731, 3758, 3784 , 3811, 3838, 3866, 3894, 3923, 3951, 3981, 
      4010, 4040, 4071 , 4102, 4133, 4165, 4197, 4230, 4263, 4297, 
      4331, 4366, 4401 , 4437, 4473, 4510, 4548, 4586, 4625, 4664, 
      4704, 4745, 4786 , 4828, 4871, 4915, 4959, 5004, 5050, 5096, 
      5144, 5192, 5242 , 5292, 5343, 5395, 5448, 5502, 5557, 5613, 
      5671, 5729, 5789 , 5850, 5912, 5975, 6040, 6106, 6174, 6243, 
      6313, 6385, 6459 , 6534, 6612, 6690, 6771, 6854, 6939, 7026, 
      7115, 7206, 7299 , 7395, 7494, 7595, 7699, 7805, 7915, 8028, 
      8144, 8263, 8385 , 8512, 8642, 8776, 8914, 9057, 9204, 9356, 
      9514, 9676, 9844 , 10018, 10198, 10384, 10578, 10779, 10987, 11204, 
      11429, 11663, 11908 , 12162, 12428, 12705, 12995, 13298, 13616, 13949, 
      14299, 14667, 15054 , 15462, 15893, 16348, 16830, 17341, 17885, 18463, 
      19079, 19738, 20444 , 21203, 22019, 22901, 23857, 24895, 26028, 27268, 
      28633, 30141, 31816 , 33689, 35795, 38182, 40910, 44058, 47730, 52069, 
      57276, 63640, 71594 , 81820, 95453, 114536, 143155, 190839, 286151, 571653, 
      251231998, -574267, -286805 , -191129, -143318, -114641, -95525, -81873, -71635, -63672, 
      -57303, -52091, -47748 , -44073, -40924, -38194, -35805, -33698, -31824, -30148, 
      -28639, -27274, -26033 , -24900, -23861, -22905, -22023, -21206, -20448, -19741, 
      -19082, -18465, -17887 , -17344, -16833, -16350, -15895, -15464, -15056, -14669, 
      -14301, -13951, -13618 , -13300, -12996, -12706, -12429, -12163, -11909, -11665, 
      -11430, -11205, -10988 , -10780, -10579, -10385, -10199, -10019, -9845, -9677, 
      -9514, -9357, -9205 , -9058, -8915, -8777, -8643, -8512, -8386, -8263, 
      -8144, -8028, -7916 , -7806, -7699, -7595, -7494, -7396, -7300, -7206, 
      -7115, -7026, -6939 , -6854, -6772, -6691, -6612, -6535, -6459, -6386, 
      -6313, -6243, -6174 , -6106, -6040, -5975, -5912, -5850, -5789, -5729, 
      -5671, -5614, -5557 , -5502, -5448, -5395, -5343, -5292, -5242, -5193, 
      -5144, -5097, -5050 , -5004, -4959, -4915, -4871, -4828, -4786, -4745, 
      -4704, -4664, -4625 , -4586, -4548, -4510, -4473, -4437, -4401, -4366, 
      -4331, -4297, -4263 , -4230, -4197, -4165, -4133, -4102, -4071, -4040, 
      -4010, -3981, -3952 , -3923, -3894, -3866, -3839, -3811, -3784, -3758, 
      -3732, -3706, -3680 , -3655, -3630, -3605, -3581, -3557, -3533, -3510, 
      -3487, -3464, -3442 , -3419, -3397, -3375, -3354, -3333, -3312, -3291, 
      -3270, -3250, -3230 , -3210, -3191, -3171, -3152, -3133, -3114, -3096, 
      -3077, -3059, -3041 , -3023, -3006, -2988, -2971, -2954, -2937, -2920, 
      -2904, -2887, -2871 , -2855, -2839, -2823, -2808, -2792, -2777, -2762, 
      -2747, -2732, -2717 , -2703, -2688, -2674, -2660, -2646, -2632, -2618, 
      -2605, -2591, -2578 , -2564, -2551, -2538, -2525, -2512, -2500, -2487, 
      -2475, -2462, -2450 , -2438, -2426, -2414, -2402, -2390, -2378, -2367, 
      -2355, -2344, -2333 , -2322, -2310, -2299, -2288, -2278, -2267, -2256, 
      -2246, -2235, -2225 , -2214, -2204, -2194, -2184, -2174, -2164, -2154, 
      -2144, -2134, -2125 , -2115, -2106, -2096, -2087, -2077, -2068, -2059, 
      -2050, -2041, -2032 , -2023, -2014, -2005, -1996, -1988, -1979, -1971, 
      -1962, -1954, -1945 , -1937, -1929, -1921, -1912, -1904, -1896, -1888, 
      -1880, -1872, -1865 , -1857, -1849, -1841, -1834, -1826, -1819, -1811, 
      -1804, -1796, -1789 , -1781, -1774, -1767, -1760, -1753, -1746, -1739, 
      -1732, -1725, -1718 , -1711, -1704, -1697, -1690, -1684, -1677, -1670, 
      -1664, -1657, -1651 , -1644, -1638, -1631, -1625, -1619, -1612, -1606, 
      -1600, -1594, -1587 , -1581, -1575, -1569, -1563, -1557, -1551, -1545, 
      -1539, -1534, -1528 , -1522, -1516, -1510, -1505, -1499, -1493, -1488, 
      -1482, -1477, -1471 , -1465, -1460, -1455, -1449, -1444, -1438, -1433, 
      -1428, -1422, -1417 , -1412, -1407, -1401, -1396, -1391, -1386, -1381, 
      -1376, -1371, -1366 , -1361, -1356, -1351, -1346, -1341, -1336, -1331, 
      -1327, -1322, -1317 , -1312, -1307, -1303, -1298, -1293, -1289, -1284, 
      -1279, -1275, -1270 , -1266, -1261, -1257, -1252, -1248, -1243, -1239, 
      -1234, -1230, -1226 , -1221, -1217, -1213, -1208, -1204, -1200, -1195, 
      -1191, -1187, -1183 , -1179, -1175, -1170, -1166, -1162, -1158, -1154, 
      -1150, -1146, -1142 , -1138, -1134, -1130, -1126, -1122, -1118, -1114, 
      -1110, -1106, -1102 , -1098, -1095, -1091, -1087, -1083, -1079, -1076, 
      -1072, -1068, -1064 , -1061, -1057, -1053, -1050, -1046, -1042, -1039, 
      -1035, -1031, -1028 , -1024, -1021, -1017, -1014, -1010, -1007, -1003, 
      -1000, -996, -993 , -989, -986, -982, -979, -975, -972, -969, 
      -965, -962, -958 , -955, -952, -948, -945, -942, -939, -935, 
      -932, -929, -926 , -922, -919, -916, -913, -909, -906, -903, 
      -900, -897, -894 , -890, -887, -884, -881, -878, -875, -872, 
      -869, -866, -863 , -860, -857, -854, -851, -848, -845, -842, 
      -839, -836, -833 , -830, -827, -824, -821, -818, -815, -812, 
      -809, -806, -804 , -801, -798, -795, -792, -789, -786, -784, 
      -781, -778, -775 , -772, -770, -767, -764, -761, -759, -756, 
      -753, -750, -748 , -745, -742, -739, -737, -734, -731, -729, 
      -726, -723, -721 , -718, -715, -713, -710, -708, -705, -702, 
      -700, -697, -695 , -692, -689, -687, -684, -682, -679, -677, 
      -674, -671, -669 , -666, -664, -661, -659, -656, -654, -651, 
      -649, -646, -644 , -641, -639, -637, -634, -632, -629, -627, 
      -624, -622, -620 , -617, -615, -612, -610, -608, -605, -603, 
      -600, -598, -596 , -593, -591, -589, -586, -584, -582, -579, 
      -577, -575, -572 , -570, -568, -565, -563, -561, -558, -556, 
      -554, -552, -549 , -547, -545, -542, -540, -538, -536, -533, 
      -531, -529, -527 , -525, -522, -520, -518, -516, -513, -511, 
      -509, -507, -505 , -502, -500, -498, -496, -494, -492, -489, 
      -487, -485, -483 , -481, -479, -476, -474, -472, -470, -468, 
      -466, -464, -462 , -459, -457, -455, -453, -451, -449, -447, 
      -445, -443, -441 , -438, -436, -434, -432, -430, -428, -426, 
      -424, -422, -420 , -418, -416, -414, -412, -410, -408, -406, 
      -404, -402, -399 , -397, -395, -393, -391, -389, -387, -385, 
      -383, -381, -379 , -377, -375, -373, -371, -369, -367, -365, 
      -363, -362, -360 , -358, -356, -354, -352, -350, -348, -346, 
      -344, -342, -340 , -338, -336, -334, -332, -330, -328, -326, 
      -324, -322, -321 , -319, -317, -315, -313, -311, -309, -307, 
      -305, -303, -301 , -300, -298, -296, -294, -292, -290, -288, 
      -286, -284, -282 , -281, -279, -277, -275, -273, -271, -269, 
      -267, -266, -264 , -262, -260, -258, -256, -254, -253, -251, 
      -249, -247, -245 , -243, -241, -240, -238, -236, -234, -232, 
      -230, -229, -227 , -225, -223, -221, -219, -218, -216, -214, 
      -212, -210, -208 , -207, -205, -203, -201, -199, -198, -196, 
      -194, -192, -190 , -188, -187, -185, -183, -181, -179, -178, 
      -176, -174, -172 , -170, -169, -167, -165, -163, -161, -160, 
      -158, -156, -154 , -153, -151, -149, -147, -145, -144, -142, 
      -140, -138, -136 , -135, -133, -131, -129, -128, -126, -124, 
      -122, -121, -119 , -117, -115, -113, -112, -110, -108, -106, 
      -105, -103, -101 , -99, -98, -96, -94, -92, -91, -89, 
      -87, -85, -83 , -82, -80, -78, -76, -75, -73, -71, 
      -69, -68, -66 , -64, -62, -61, -59, -57, -55, -54, 
      -52, -50, -48 , -47, -45, -43, -41, -40, -38, -36, 
      -34, -33, -31 , -29, -27, -26, -24, -22, -20, -19, 
      -17, -15, -13 , -12, -10, -8, -6, -5, -3, -1
};


void
r_draw_gradient_axial (render_ctx *ctx, int scol, int ecol, int angle)
{
	GdkColor cols[48];
	int steps = 48, w = ctx->pix_w, h = ctx->pix_h;
	unsigned int sr,sg,sb, er,eg,eb, dr,dg,db;
	int U, px, py, distance, n = 0;
	int x, y;
	double x2, y2, mm;
	int axiald;
	guchar *rgb_buf, *pos;

	y2 = (tantable[angle]* (w/2) /1000) - (h/2);
	mm = (2 * y2 + h) * (2 * y2 + h) + w * w;
	if (angle == 0) {
		axiald = h/2;
	} else if (angle == 900) {
		axiald = w/2;
	} else {
		x2 = sqrt(10*w*h/tantable[angle]);
		axiald = ((w-x2)*tantable[angle]/1000);
	}

	sr = (scol >> 16) & 0xFF;
	sg = (scol >> 8) & 0xFF;
	sb = scol & 0xFF;
	er = (ecol >> 16) & 0xFF;
	eg = (ecol >> 8) & 0xFF;
	eb = ecol & 0xFF;

	dr = (er - sr) * 100 / (steps - 1);
	dg = (eg - sg) * 100 / (steps - 1);
	db = (eb - sb) * 100 / (steps - 1);
	for (x = 0; x < steps; x++) {
		cols[x].red = sr + (dr * (steps - x - 1) / 100);
		cols[x].green = sg + (dg * (steps - x - 1) / 100);
		cols[x].blue = sb + (db * (steps - x - 1) / 100);
//		gdk_colormap_alloc_color (gdk_drawable_get_colormap (ctx->d), &cols[x], FALSE, TRUE);
	}

	rgb_buf = malloc (w * h * 3);
	pos = rgb_buf;
	for (y = 0; y < h; y++) {
		int tempy = (h + y2 - y) * (2 * y2 + h);
		for (x = 0; x < w; x++) {
			U = (x * w) + tempy;
			if (U > 0 && U < mm) {
				px = U * w / mm;
				py = (U * (h + 2 * y2)) / mm - (h + y2);
				distance = sqrt ((px - x) * (px - x) + (py + y) * (py + y));
				if (axiald == 0)
					n = 0;
				else
					n = (distance * 1024 / axiald) * (steps - 1) >> 10;
				if (n > (steps - 1)) n = steps - 1;
			}
//			gdk_gc_set_rgb_fg_color (ctx->gc, &cols[steps-n-1]);
//			gdk_draw_point (ctx->d, ctx->gc, x, y);
			*pos++ = cols[steps-n-1].red;
			*pos++ = cols[steps-n-1].green;
			*pos++ = cols[steps-n-1].blue;
		}
	}
	gdk_draw_rgb_image (ctx->d, ctx->gc, 0, 0, w, h, GDK_RGB_DITHER_NONE, rgb_buf, w * 3);
	free (rgb_buf);
}

void
r_draw_gradient_linear (render_ctx *ctx, int scol, int ecol, int angle)
{
	GdkColor cols[256];
	int steps = 256, w = ctx->pix_w, h = ctx->pix_h;
	unsigned int sr,sg,sb, er,eg,eb, dr,dg,db;
	unsigned int U, px, py, n = 0;
	double distance;
	int x, y;
	guchar *rgb_buf, *pos;

	sr = (scol >> 16) & 0xFF;
	sg = (scol >> 8) & 0xFF;
	sb = scol & 0xFF;
	er = (ecol >> 16) & 0xFF;
	eg = (ecol >> 8) & 0xFF;
	eb = ecol & 0xFF;

	dr = (er - sr) * 100 / (steps - 1);
	dg = (eg - sg) * 100 / (steps - 1);
	db = (eb - sb) * 100 / (steps - 1);
	for (x = 0; x < steps; x++) {
		cols[x].red = sr + (dr * (steps - x - 1) / 100);
		cols[x].green = sg + (dg * (steps - x - 1) / 100);
		cols[x].blue = sb + (db * (steps - x - 1) / 100);
	}

	rgb_buf = malloc (w * h * 3);
	pos = rgb_buf;

	if ((angle % 900) != 0) {
		unsigned int x2 = 512*h/tantable[angle];
		unsigned int y2 = w*512/tantable[angle];
		unsigned int x3 = w + x2;
		unsigned int y3 = h + y2;
		unsigned int xy2 = (x3*x3+y3*y3);
		unsigned int axiald;

		axiald = sqrt (w*w + h*h);

		for (y=0; y < h; y++) {
			int tempy = (h-y)*y3;
			for (x = 0; x < w; x++) {
				U = (x+x2)*x3+tempy;
				if ( (U > 0) && (U < xy2)) {
					px = (U*(x2+w)/xy2)-x2;
					py = (U*(y2+h)/xy2)-h;
					distance = sqrt(((px-x)*(px-x))+((py+y)*(py+y)));
					n = ((int)((distance*1024/axiald)*(steps-1))) >> 10;
					if (n > (steps-1)) n = (steps-1);
				}
				*pos++ = cols[steps-n-1].red;
				*pos++ = cols[steps-n-1].green;
				*pos++ = cols[steps-n-1].blue;
			}
		}

	} else {
		for (y = 0; y < h; y++) {
			for (x = 0; x < w; x++) {
				if (angle == 900) {
					n = (((w-x)*1024/w)*(steps-1)>>10);
				} else if (angle == 1800) {
					n = (((h-y)*1024/h)*(steps-1)>>10);
				} else if (angle == 2700) {
					n = ((x*1024/w)*(steps-1)>>10);
				} else {
					n = ((y*1024/h)*(steps-1)>>10);
				}
   				*pos++ = cols[steps-n-1].red;
				*pos++ = cols[steps-n-1].green;
				*pos++ = cols[steps-n-1].blue;
			}
		}
	}

	gdk_draw_rgb_image (ctx->d, ctx->gc, 0, 0, w, h, GDK_RGB_DITHER_NONE, rgb_buf, w * 3);
	free (rgb_buf);
}
